/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.repo.db;

import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.filteredresources.FilteredResourcesAddon;
import org.artifactory.api.repo.exception.RepoRejectException;
import org.artifactory.checksum.ChecksumsInfo;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RepoBaseDescriptor;
import org.artifactory.fs.FileInfo;
import org.artifactory.fs.RepoResource;
import org.artifactory.fs.ZipEntryRepoResource;
import org.artifactory.io.ZipResourceStreamHandle;
import org.artifactory.model.common.RepoPathImpl;
import org.artifactory.model.xstream.fs.FileInfoImpl;
import org.artifactory.repo.RepoPathFactory;
import org.artifactory.repo.SaveResourceContext;
import org.artifactory.request.Request;
import org.artifactory.request.RequestContext;
import org.artifactory.resource.ResourceStreamHandle;
import org.artifactory.resource.ResourceWithChecksumStreamHandle;
import org.artifactory.storage.StorageException;
import org.artifactory.storage.db.fs.model.DbFsFile;
import org.artifactory.storage.fs.service.StatsService;
import org.artifactory.util.RepoPathUtils;
import org.jetbrains.annotations.NotNull;
import org.springframework.test.util.ReflectionTestUtils;
import org.testng.annotations.Test;

import java.io.IOException;
import java.io.InputStream;

import static org.mockito.Mockito.any;
import static org.mockito.Mockito.doReturn;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertTrue;

/**
 * Unit tests for the save resource process.
 *
 * @author Yossi Shaul
 * @see org.artifactory.repo.db.DbStoringRepoMixin#saveResource(org.artifactory.repo.SaveResourceContext)
 */
@Test
public class DbStoringRepoMixinSaveResourceTest {

    @Test(expectedExceptions = StorageException.class, expectedExceptionsMessageRegExp = ".*parent.*")
    public void failToSaveRootFolder() throws IOException, RepoRejectException {
        DbStoringRepoMixin<RepoBaseDescriptor> repo = createStoringRepo();
        RepoResource repoResourceMock = mock(RepoResource.class);
        when(repoResourceMock.getRepoPath()).thenReturn(RepoPathUtils.repoRootPath("repo"));
        repo.saveResource(new SaveResourceContext.Builder(repoResourceMock, (InputStream) null).build());
    }

    private DbStoringRepoMixin<RepoBaseDescriptor> createStoringRepo() {
        LocalRepoDescriptor repoDescriptor = new LocalRepoDescriptor();
        repoDescriptor.setKey("repo");
        return spy(new DbStoringRepoMixin<>(repoDescriptor, null));
    }

    @Test
    public void testGetResourceStreamHandleSimpleStream() throws IOException {
        DbStoringRepoMixin<RepoBaseDescriptor> repo = createStoringRepo();
        RequestContext requestContext = mock(RequestContext.class);
        RepoResource res = mock(RepoResource.class);
        when(res.getRepoPath()).thenReturn(RepoPathFactory.create("repo", "path"));
        ChecksumsInfo checksumInfo = prepareChecksums(repo);

        prepareRequest(requestContext, false);
        prepareStats(repo);
        ResourceStreamHandle resourceStreamHandle = repo.getResourceStreamHandle(requestContext, res);
        assertTrue(resourceStreamHandle instanceof ResourceWithChecksumStreamHandle);
        assertTrue(((ResourceWithChecksumStreamHandle) resourceStreamHandle).getChecksumsInfo().isPresent());
        assertEquals(((ResourceWithChecksumStreamHandle) resourceStreamHandle).getChecksumsInfo().get(), checksumInfo);
    }

    @Test
    public void testGetResourceStreamHandleZipStream() throws IOException {
        DbStoringRepoMixin<RepoBaseDescriptor> repo = createStoringRepo();
        RequestContext requestContext = mock(RequestContext.class);
        ZipEntryRepoResource res = mock(ZipEntryRepoResource.class);
        when(res.getRepoPath()).thenReturn(RepoPathFactory.create("repo", "path"));
        FileInfo fileInfo = spy(new FileInfoImpl(RepoPathFactory.create("repo", "path")));
        ChecksumsInfo checksumInfo = new ChecksumsInfo();
        when(fileInfo.getChecksumsInfo()).thenReturn(checksumInfo);

        when(res.getInfo()).thenReturn(fileInfo);
        prepareRequest(requestContext, true);
        prepareStats(repo);
        prepareVfsFile(repo, fileInfo);
        AddonsManager addonsManager = prepareAddonManager(repo);
        FilteredResourcesAddon filteredResourcesAddon = mock(FilteredResourcesAddon.class);
        ZipResourceStreamHandle zipResourceStreamHandle = mock(ZipResourceStreamHandle.class);
        when(filteredResourcesAddon.getZipResourceHandle(res, null, checksumInfo)).thenReturn(zipResourceStreamHandle);
        when(addonsManager.addonByType(FilteredResourcesAddon.class)).thenReturn(filteredResourcesAddon);

        ResourceStreamHandle resourceStreamHandle = repo.getResourceStreamHandle(requestContext, res);
        assertTrue(resourceStreamHandle instanceof ResourceWithChecksumStreamHandle);
        assertEquals(resourceStreamHandle, zipResourceStreamHandle);
    }

    private void prepareStats(DbStoringRepoMixin<RepoBaseDescriptor> repo) {
        StatsService statsService = mock(StatsService.class);
        ReflectionTestUtils.setField(repo, "statsService", statsService);
    }

    private AddonsManager prepareAddonManager(DbStoringRepoMixin<RepoBaseDescriptor> repo) {
        AddonsManager addonsManager = mock(AddonsManager.class);
        ReflectionTestUtils.setField(repo, "addonsManager", addonsManager);
        return addonsManager;
    }

    @NotNull
    private ChecksumsInfo prepareChecksums(DbStoringRepoMixin<RepoBaseDescriptor> repo) {
        FileInfo fileInfo = spy(new FileInfoImpl(RepoPathFactory.create("repo", "path")));
        ChecksumsInfo checksumInfo = new ChecksumsInfo();
        when(fileInfo.getChecksumsInfo()).thenReturn(checksumInfo);
        prepareVfsFile(repo, fileInfo);
        return checksumInfo;
    }

    private void prepareVfsFile(DbStoringRepoMixin<RepoBaseDescriptor> repo, FileInfo fileInfo) {
        DbFsFile vfsFile = mock(DbFsFile.class);
        when(vfsFile.getInfo()).thenReturn(fileInfo);
        doReturn(vfsFile).when(repo).getImmutableFile(any(RepoPathImpl.class));
    }

    private void prepareRequest(RequestContext requestContext, boolean isZip) {
        Request request = mock(Request.class);
        when(requestContext.getRequest()).thenReturn(request);
        when(request.isZipResourceRequest()).thenReturn(isZip);
    }
}