/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.storage.db.build.itest.dao;

import com.google.common.collect.ImmutableSet;
import com.google.common.collect.ImmutableSortedSet;
import com.google.common.collect.UnmodifiableIterator;
import org.artifactory.addon.build.BuildAddon;
import org.artifactory.api.build.GeneralBuild;
import org.artifactory.api.build.ModuleArtifact;
import org.artifactory.api.build.ModuleDependency;
import org.artifactory.api.build.PublishedModule;
import org.artifactory.api.build.model.diff.BuildParams;
import org.artifactory.api.rest.build.ContinueBuildFilter;
import org.artifactory.build.BuildId;
import org.artifactory.build.BuildNameAndRepo;
import org.artifactory.checksum.ChecksumType;
import org.artifactory.storage.db.binstore.dao.BinariesDao;
import org.artifactory.storage.db.build.entity.BuildEntity;
import org.artifactory.storage.db.build.entity.BuildIdEntity;
import org.artifactory.storage.db.build.entity.BuildPromotionStatus;
import org.artifactory.storage.db.build.entity.BuildProperty;
import org.artifactory.storage.db.build.service.BuildIdImpl;
import org.artifactory.storage.db.fs.dao.NodesDao;
import org.artifactory.storage.db.fs.dao.PropertiesDao;
import org.fest.assertions.Assertions;
import org.jfrog.build.api.release.PromotionStatus;
import org.joda.time.DateTime;
import org.springframework.beans.factory.annotation.Autowired;
import org.testng.annotations.AfterClass;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.Test;

import java.sql.SQLException;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertNotNull;
import static org.testng.Assert.assertNull;
import static org.testng.Assert.assertTrue;
import static org.testng.Assert.fail;

/**
 * Date: 10/31/12
 * Time: 11:03 AM
 *
 * @author freds
 */
@Test
public class BuildsDaoTest extends BuildsDaoBaseTest {

    private static final String ALTERNATIVE_BUILD_INFO_REPO = "prjx-build-info";
    private static final String NOT_EXISTING_BUILD_REPO = "not_existing_build_repo";

    @Autowired
    private BinariesDao binariesDao;
    @Autowired
    private NodesDao nodesDao;
    @Autowired
    private PropertiesDao propertiesDao;

    @BeforeClass
    public void setup() throws SQLException {
        buildsDao.deleteAllBuilds(false);
        importSql("/sql/builds_in_two_repos.sql");
    }

    public void testReadBuildsById() throws SQLException {
        for (long i = 1L; i <= 6L; i++) {
            assertAnyBuild(buildsDao.getBuild(i));
        }
    }

    /*
    public void testReplace() {
        final String buildName = "??na**me";
        assertEquals(buildName.replaceAll("\\?", "_").replaceAll("\\*", "%"), "__na%%me");
        assertEquals(buildName.replace("?", "_").replace("*", "%"), "__na%%me");
    }
    */

    public void testFindLatestBuildIdsByNameCaseInsensitive() throws SQLException {
        DateTime buildDate = DateTime.now().minusHours(1);
        BuildEntity myBuild_1 =
                createBuild(System.nanoTime(), "myBuild", "1", buildDate.getMillis(), BuildAddon.BUILD_INFO_REPO_NAME);
        saveBuild(myBuild_1);

        BuildEntity mybuild_1 =
                createBuild(System.nanoTime(), "mybuild", "1", buildDate.getMillis(), BuildAddon.BUILD_INFO_REPO_NAME);
        saveBuild(mybuild_1);

        BuildEntity mybuild_2 = createBuild(System.nanoTime(), "mybuild", "2", buildDate.plusMinutes(2).getMillis(),
                BuildAddon.BUILD_INFO_REPO_NAME);
        saveBuild(mybuild_2);

        List<BuildIdEntity> latestBuildIds =
                buildsDao.findLatestBuildIdsByNameCaseInsensitive(null, "myb", "", "", "date", "desc");

        assert latestBuildIds.size() == 2;
        assertEquals(latestBuildIds.get(0).getBuildName(), mybuild_2.getBuildName());
        assertEquals(latestBuildIds.get(0).getBuildNumber(), mybuild_2.getBuildNumber());
        assertEquals(latestBuildIds.get(1).getBuildName(), myBuild_1.getBuildName());
    }

    @Test(dependsOnMethods = "testDeleteBuilds")
    public void testReadBuildsByNameAndRepo() throws SQLException {
        List<Long> buildIds = buildsDao.findBuildIds("ba", BuildAddon.BUILD_INFO_REPO_NAME);
        assertEquals(buildIds.size(), 3);
        for (Long buildId : buildIds) {
            assertTrue(buildId == 1L || buildId == 3L || buildId == 5L, "Build id " + buildId + " should not be here!");
        }
        buildIds = buildsDao.findBuildIds("bb", ALTERNATIVE_BUILD_INFO_REPO);
        assertEquals(buildIds.size(), 2);
        for (Long buildId : buildIds) {
            assertTrue(buildId == 2L || buildId == 4L, "Build id " + buildId + " should not be here!");
        }
        buildIds = buildsDao.findBuildIds("bb", BuildAddon.BUILD_INFO_REPO_NAME);
        assertEquals(buildIds.size(), 1);
    }

    @Test(dependsOnMethods = "testDeleteBuilds")
    public void testReadBuildsByNameNumberRepo() throws SQLException {
        List<Long> buildIds = buildsDao.findBuildIds("ba", "1", BuildAddon.BUILD_INFO_REPO_NAME);
        assertEquals(buildIds.size(), 1);
        buildIds = buildsDao.findBuildIds("bb", "1", ALTERNATIVE_BUILD_INFO_REPO);
        assertEquals(buildIds.size(), 1);
        buildIds = buildsDao.findBuildIds("bb", "1", BuildAddon.BUILD_INFO_REPO_NAME);
        assertEquals(buildIds.size(), 0);
    }

    @Test(dependsOnMethods = "testDeleteBuilds")
    public void testGetAllBuildNames() throws SQLException {
        List<BuildNameAndRepo> buildNames = buildsDao.getAllBuildNames(null);
        assertEquals(buildNames.size(), 8);
        buildNames = buildsDao.getAllBuildNames(BuildAddon.BUILD_INFO_REPO_NAME);
        assertEquals(buildNames.size(), 6);
        buildNames = buildsDao.getAllBuildNames(ALTERNATIVE_BUILD_INFO_REPO);
        assertEquals(buildNames.size(), 2);
    }

    @Test(priority = 100)
    public void testGetLatestBuildIds() throws SQLException {
        buildArtifactsDao.deleteAllBuildArtifacts();
        buildDependenciesDao.deleteAllBuildDependencies();
        buildModulesDao.deleteAllBuildModules();
        buildsDao.deleteAllBuilds(false);
        importSql("/sql/builds.sql");
        ContinueBuildFilter continueBuildFilter = new ContinueBuildFilter();
        List<BuildIdEntity> latestBuildIds = buildsDao.getLatestBuildIds(continueBuildFilter);
        assertEquals(latestBuildIds.size(), 2);
        Map<String, BuildIdEntity> builds = latestBuildIds.stream()
                .collect(Collectors.toMap(BuildIdEntity::getBuildName, build -> build));
        assertEquals(builds.get("bb").getBuildId(), 4);
        assertEquals(builds.get("bb").getBuildNumber(), "2");
        assertEquals(builds.get("bb").getBuildDate(), 1349003000000L);
        assertEquals(builds.get("ba").getBuildId(), 5);
        assertEquals(builds.get("ba").getBuildNumber(), "3");
        assertEquals(builds.get("ba").getBuildDate(), 1349004000000L);
    }

    @Test(dependsOnMethods = "testDeleteBuilds")
    public void testGetLatestBuildIdsByBuildRepoOnFilter() {
        ContinueBuildFilter continueBuildFilter = new ContinueBuildFilter();
        continueBuildFilter.setBuildRepo("prjx-build-info");
        List<BuildIdEntity> latestBuildIds = buildsDao.getLatestBuildIds(continueBuildFilter);
        assertEquals(latestBuildIds.size(), 2);
        Map<String, BuildIdEntity> builds = latestBuildIds.stream()
                .collect(Collectors.toMap(BuildIdEntity::getBuildName, build -> build));

        assertEquals(builds.get("bb").getBuildId(), 4);
        assertEquals(builds.get("bb").getBuildNumber(), "2");
        assertEquals(builds.get("bb").getBuildDate(), 1349003000000L);

        assertEquals(builds.get("cc").getBuildId(), 8);
        assertEquals(builds.get("cc").getBuildNumber(), "1");
        assertEquals(builds.get("cc").getBuildDate(), 1349006000000L);
    }

    @Test(dependsOnMethods = "testDeleteBuilds")
    public void testGetLatestBuildIdsByBuildRepoOnBuildId() {
        ContinueBuildFilter continueBuildFilter = new ContinueBuildFilter();
        BuildId buildId = new BuildIdImpl(2L, "bb", "1", 1349001000000L, "prjx-build-info");
        continueBuildFilter.setContinueBuildId(buildId);
        List<BuildIdEntity> latestBuildIds = buildsDao.getLatestBuildIds(continueBuildFilter);
        assertEquals(latestBuildIds.size(), 2);
        Map<String, BuildIdEntity> builds = latestBuildIds.stream()
                .collect(Collectors.toMap(BuildIdEntity::getBuildName, build -> build));
        assertEquals(builds.get("bb").getBuildId(), 4);
        assertEquals(builds.get("bb").getBuildNumber(), "2");
        assertEquals(builds.get("bb").getBuildDate(), 1349003000000L);

        assertEquals(builds.get("cc").getBuildId(), 8);
        assertEquals(builds.get("cc").getBuildNumber(), "1");
        assertEquals(builds.get("cc").getBuildDate(), 1349006000000L);
    }

    public void testReadBuildsByFullPath() throws SQLException {
        assertBuild1(buildsDao.findBuild("ba", "1", 1349000000000L));
        assertBuild2(buildsDao.findBuild("bb", "1", 1349001000000L));
        assertBuild3(buildsDao.findBuild("ba", "2", 1349002000000L));
        assertBuild4(buildsDao.findBuild("bb", "2", 1349003000000L));
        assertBuild5(buildsDao.findBuild("ba", "3", 1349004000000L));
    }

    public void testSimpleCreateBuild() throws SQLException {
        BuildEntity build11 = createBuild11();
        assertEquals(saveBuild(build11), 1);
        assertTrue(build11.isIdentical(buildsDao.getBuild(11L)));
    }

    public void testCreateBuildWithProps() throws SQLException {
        BuildEntity c2 = createBuild12();
        assertEquals(saveBuild(c2), 3);
        assertTrue(c2.isIdentical(buildsDao.getBuild(12L)));
    }

    public void testCreateBuildWithPromotions() throws SQLException {
        BuildEntity c2 = createBuild13();
        assertEquals(saveBuild(c2), 3);
        assertTrue(c2.isIdentical(buildsDao.getBuild(13L)));
    }

    public void testCreateBuildWithPropsAndPromotions() throws SQLException {
        BuildEntity c2 = createBuild14();
        assertEquals(saveBuild(c2), 5);
        assertTrue(c2.isIdentical(buildsDao.getBuild(14L)));
    }

    public void testCreateBuildWithLongModuleProps() throws SQLException {
        BuildEntity c5 = createBuild15();
        assertEquals(saveBuild(c5), 2);
        BuildEntity build = buildsDao.getBuild(15L);
        build.getProperties().forEach(buildProperty ->
                assertTrue(buildProperty.getPropValue().length() <= buildsDao.getDbIndexedValueMaxSize(2048)));
    }

    public void testCreateBuildSameNameAndNumber() throws SQLException {
        long now = System.currentTimeMillis();
        BuildEntity otherBa2 = new BuildEntity(103L, "ba", "2", 1349111111111L, null, now, "me", 0L, null,
                BuildAddon.BUILD_INFO_REPO_NAME);
        otherBa2.setProperties(new HashSet<>());
        otherBa2.setPromotions(new HashSet<>());
        assertEquals(saveBuild(otherBa2), 1);
        assertTrue(otherBa2.isIdentical(buildsDao.getBuild(103L)));
    }

    @Test(dependsOnMethods = "testCreateBuildSameNameAndNumber")
    public void testFindLatestBuild() throws SQLException {
        assertEquals(buildsDao.findLatestBuild("ba", "1", BuildAddon.BUILD_INFO_REPO_NAME).getStartedDate().getTime(),
                1349000000000L);
        assertEquals(buildsDao.findLatestBuild("bb", "1", ALTERNATIVE_BUILD_INFO_REPO).getStartedDate().getTime(),
                1349001000000L);
        assertEquals(buildsDao.findLatestBuild("ba", "2", BuildAddon.BUILD_INFO_REPO_NAME).getStartedDate().getTime(),
                1349111111111L);
        assertEquals(buildsDao.findLatestBuild("bb", "2", ALTERNATIVE_BUILD_INFO_REPO).getStartedDate().getTime(),
                1349003000000L);
        assertEquals(buildsDao.findLatestBuild("ba", "3", BuildAddon.BUILD_INFO_REPO_NAME).getStartedDate().getTime(),
                1349004000000L);
        assertEquals(buildsDao.findLatestBuild("ba", "3", BuildAddon.BUILD_INFO_REPO_NAME).getBuildRepo(),
                BuildAddon.BUILD_INFO_REPO_NAME);
        assertEquals(buildsDao.findLatestBuild("bb", "2", ALTERNATIVE_BUILD_INFO_REPO).getBuildRepo(),
                ALTERNATIVE_BUILD_INFO_REPO);
        assertNotNull(buildsDao.findLatestBuild("bb", "3", BuildAddon.BUILD_INFO_REPO_NAME));
    }

    @Test(dependsOnMethods = "testCreateBuildSameNameAndNumber")
    public void testFindLatestBuildIdsByStoringRepo() throws SQLException {
        // When repo param is specified we expect to get just the build names from that specific build-info repo
        List<BuildIdEntity> latestBuildIds =
                buildsDao.findLatestBuildIdsByNameCaseInsensitive(ALTERNATIVE_BUILD_INFO_REPO, "bb", "", "", "date",
                        "desc");
        assertEquals(latestBuildIds.size(), 1);
        assertEquals(latestBuildIds.get(0).getBuildName(), "bb");
        assertEquals(latestBuildIds.get(0).getBuildNumber(), "2");

        latestBuildIds =
                buildsDao.findLatestBuildIdsByNameCaseInsensitive(BuildAddon.BUILD_INFO_REPO_NAME, "bb", "", "", "date",
                        "desc");
        assertEquals(latestBuildIds.size(), 1);
        assertEquals(latestBuildIds.get(0).getBuildName(), "bb");
        assertEquals(latestBuildIds.get(0).getBuildNumber(), "3");

        // When repo param is null we expect to get all builds - like before the repo column was added
        latestBuildIds = buildsDao.findLatestBuildIdsByNameCaseInsensitive(null, "bb", "", "", "date", "desc");
        assertEquals(latestBuildIds.size(), 2);
        assertEquals(latestBuildIds.get(0).getBuildName(), "bb");
        assertEquals(latestBuildIds.get(0).getBuildNumber(), "3");
        assertEquals(latestBuildIds.get(1).getBuildName(), "bb");
        assertEquals(latestBuildIds.get(1).getBuildNumber(), "2");
    }

    @Test(dependsOnMethods = "testFindLatestBuildIdsByStoringRepo")
    public void testGetBuildsCountFromSpecificRepo() throws SQLException {
        assertEquals(buildsDao.getBuildsCountFromSpecificRepo(ALTERNATIVE_BUILD_INFO_REPO), 3);
        assertEquals(buildsDao.getBuildsCountFromSpecificRepo(BuildAddon.BUILD_INFO_REPO_NAME), 14);
    }

    @Test(dependsOnMethods = "testFindLatestBuildIdsByStoringRepo")
    public void testGetUniqueBuildsCountFromSpecificRepo() throws SQLException {
        assertEquals(buildsDao.getUniqueBuildsCountFromSpecificRepo(ALTERNATIVE_BUILD_INFO_REPO), 2);
        assertEquals(buildsDao.getUniqueBuildsCountFromSpecificRepo(BuildAddon.BUILD_INFO_REPO_NAME), 8);
    }

    @Test(dependsOnMethods = "testFindLatestBuildIdsByStoringRepo")
    public void testGetTotalBuildsCount() throws SQLException {
        assertEquals(buildsDao.getTotalBuildsCount(), 17);
    }

    @Test(dependsOnMethods = "testGetBuildsCountFromSpecificRepo")
    public void testGetBuildStoringRepo() throws SQLException {
        assertEquals(buildsDao.getBuildStoringRepo("ba", "1", 1349000000000L), BuildAddon.BUILD_INFO_REPO_NAME);
        assertEquals(buildsDao.getBuildStoringRepo("bb", "1", 1349001000000L), ALTERNATIVE_BUILD_INFO_REPO);
        assertNull(buildsDao.getBuildStoringRepo("not_exist", "1", 1349001000000L));
    }

    @Test(dependsOnMethods = "testGetBuildsCountFromSpecificRepo")
    public void testGetBuildModule() throws SQLException {
        List<PublishedModule> publishedModules =
                buildsDao.getBuildModule("1", String.valueOf(1349000000000L), BuildAddon.BUILD_INFO_REPO_NAME);
        assertEquals(publishedModules.size(), 1);
        PublishedModule module = publishedModules.get(0);
        assertEquals(module.getNumOfArtifact(), "2");
        assertEquals(module.getNumOfDependencies(), "1");
        assertEquals(module.getId(), "ba:moda1");

        publishedModules =
                buildsDao.getBuildModule("1", String.valueOf(1349000000000L), ALTERNATIVE_BUILD_INFO_REPO);
        assertEquals(publishedModules.size(), 0);
    }

    @Test(dependsOnMethods = "testGetBuildsCountFromSpecificRepo")
    public void testGetModuleArtifact() throws SQLException {
        List<ModuleArtifact> moduleArtifacts = buildsDao
                .getModuleArtifact("ba", "1", String.valueOf(1349000000000L), "ba:moda1",
                        BuildAddon.BUILD_INFO_REPO_NAME, true);
        assertEquals(moduleArtifacts.size(), 2);
        ModuleArtifact moduleArtifact = moduleArtifacts.get(0);
        assertEquals(moduleArtifact.getName(), "ba1mod1-art1");
        assertEquals(moduleArtifact.getSha1(), "acab88fc2a043c2479a6de676a2f8179e9ea2167");
        assertEquals(moduleArtifact.getRepoKey(), "npm-local");
        assertEquals(moduleArtifact.getPath(), "test_pack1/-/test_pack-1.0.0.tgz");
        assertEquals(moduleArtifacts.get(1).getName(), "ba1mod1-art2");

        moduleArtifacts = buildsDao
                .getModuleArtifact("ba", "1", String.valueOf(1349000000000L), "ba:moda1",
                        ALTERNATIVE_BUILD_INFO_REPO, true);
        assertEquals(moduleArtifacts.size(), 0);
    }

    //live
    @Test(dependsOnMethods = "testGetBuildsCountFromSpecificRepo")
    public void testGetModuleArtifactWithRepo() throws SQLException {
        List<ModuleArtifact> moduleArtifacts = buildsDao
                .getModuleArtifact("cc", "1", String.valueOf(1349005000000L), "cc:moda1",
                        BuildAddon.BUILD_INFO_REPO_NAME, false);
        assertEquals(moduleArtifacts.size(), 1);
        ModuleArtifact moduleArtifact = moduleArtifacts.get(0);
        assertEquals(moduleArtifact.getName(), "cc2mod4-art1");
        assertEquals(moduleArtifact.getSha1(), "acab88fc2a043c2479a6de676a2f8179e9ea2167");
        assertEquals(moduleArtifact.getRepoKey(), "npm-local");
        assertEquals(moduleArtifact.getPath(), "test_pack5/-/test_pack-1.0.0.tgz");

        moduleArtifacts = buildsDao
                .getModuleArtifact("cc", "1", String.valueOf(1349006000000L), "cc:moda1",
                        ALTERNATIVE_BUILD_INFO_REPO, false);
        assertEquals(moduleArtifacts.size(), 1);
        moduleArtifact = moduleArtifacts.get(0);
        assertEquals(moduleArtifact.getName(), "cc2mod4-art1");
        assertEquals(moduleArtifact.getSha1(), "acab88fc2a043c2479a6de676a2f8179e9ea2167");
        assertEquals(moduleArtifact.getRepoKey(), "npm-local");
        assertEquals(moduleArtifact.getPath(), "test_pack4/-/test_pack-1.0.0.tgz");
    }

    @Test(dependsOnMethods = "testGetBuildsCountFromSpecificRepo")
    public void testGetModuleDependency() throws SQLException {
        List<ModuleDependency> moduleDependencies = buildsDao
                .getModuleDependency("1", String.valueOf(1349000000000L), BuildAddon.BUILD_INFO_REPO_NAME, "ba:moda1");
        assertEquals(moduleDependencies.size(), 1);
        ModuleDependency dependency = moduleDependencies.get(0);
        assertEquals(dependency.getName(), "bb1mod1-art1");
        assertEquals(dependency.getSha1(), "ccab88fc2a043c2479a6de676a2f8179e9ea2167");

        moduleDependencies = buildsDao
                .getModuleDependency("1", String.valueOf(1349000000000L), ALTERNATIVE_BUILD_INFO_REPO, "ba:moda1");
        assertEquals(moduleDependencies.size(), 0);
    }

    @Test(dependsOnMethods = "testGetBuildsCountFromSpecificRepo")
    public void testGetModuleDependencyForDiffWithPaging() {
        BuildParams buildParams =
                new BuildParams("ba:moda1", "1", "2", String.valueOf(1349002000000L), String.valueOf(1349000000000L),
                        "ba",
                        BuildAddon.BUILD_INFO_REPO_NAME);
        List<ModuleDependency> moduleDependencies = buildsDao.getModuleDependencyForDiffWithPaging(buildParams);
        assertEquals(moduleDependencies.size(), 2);

        ModuleDependency dependency1 = moduleDependencies.get(0);
        assertEquals(dependency1.getName(), "ba2mod4-art1");
        assertEquals(dependency1.getSha1(), "dcab88fc2a043c2479a6de676a2f8179e9ea2167");
        ModuleDependency dependency2 = moduleDependencies.get(1);
        assertEquals(dependency2.getName(), "bb1mod1-art1");
        assertEquals(dependency2.getSha1(), "ccab88fc2a043c2479a6de676a2f8179e9ea2167");

        buildParams.setBuildRepo(ALTERNATIVE_BUILD_INFO_REPO);
        moduleDependencies = buildsDao.getModuleDependencyForDiffWithPaging(buildParams);
        assertEquals(moduleDependencies.size(), 0);
    }

    @Test(dependsOnMethods = "testGetBuildsCountFromSpecificRepo")
    public void testGetModuleArtifactsForDiffWithPaging() {
        BuildParams buildParams =
                new BuildParams("ba:moda1", "1", "2", String.valueOf(1349002000000L), String.valueOf(1349000000000L),
                        "ba",
                        BuildAddon.BUILD_INFO_REPO_NAME);
        List<ModuleArtifact> moduleArtifacts = buildsDao.getModuleArtifactsForDiffWithPaging(buildParams, true);
        assertEquals(moduleArtifacts.size(), 6);

        buildParams.setBuildRepo(ALTERNATIVE_BUILD_INFO_REPO);
        moduleArtifacts = buildsDao.getModuleArtifactsForDiffWithPaging(buildParams, false);
        assertEquals(moduleArtifacts.size(), 0);
    }

    @Test(dependsOnMethods = "testGetBuildsCountFromSpecificRepo")
    public void testGetPrevBuildsList() {
        List<GeneralBuild> buildList =
                buildsDao.getPrevBuildsList("ba", String.valueOf(1349002000000L), BuildAddon.BUILD_INFO_REPO_NAME);
        assertEquals(buildList.size(), 1);
        assertEquals(buildList.get(0).getBuildDate(), 1349000000000L);

        buildList = buildsDao.getPrevBuildsList("ba", String.valueOf(1349002000000L), ALTERNATIVE_BUILD_INFO_REPO);
        assertEquals(buildList.size(), 0);

        buildList = buildsDao.getPrevBuildsList("ba", String.valueOf(1349000000000L), BuildAddon.BUILD_INFO_REPO_NAME);
        assertEquals(buildList.size(), 0);
    }

    @Test(dependsOnMethods = "testGetBuildsCountFromSpecificRepo")
    public void testGetLatestBuild() throws SQLException {
        BuildEntity buildEntity = buildsDao.getLatestBuild("ba", BuildAddon.BUILD_INFO_REPO_NAME);
        assertEquals(buildEntity.getBuildName(), "ba");
        assertEquals(buildEntity.getBuildNumber(), "3");
        assertEquals(buildEntity.getBuildDate(), 1349004000000L);
        assertEquals(buildEntity.getRepo(), BuildAddon.BUILD_INFO_REPO_NAME);
        assertEquals(buildEntity.getPromotions().size(), 2);
        assertEquals(buildEntity.getProperties().size(), 2);

        buildEntity = buildsDao.getLatestBuild("ba", ALTERNATIVE_BUILD_INFO_REPO);
        assertNull(buildEntity);
    }


    @Test(dependsOnMethods = {"testSimpleCreateBuild",
            "testCreateBuildWithProps",
            "testCreateBuildWithPromotions",
            "testCreateBuildWithPropsAndPromotions",
            "testFindLatestBuild"})
    public void testDeleteBuilds() throws SQLException {
        doDelete(11L, 1);
        doDelete(12L, 3);
        doDelete(13L, 3);
        doDelete(14L, 5);
        doDelete(103L, 1);
    }

    // Disable since index too big => Need to find another way for this
    @Test(enabled = false, expectedExceptions = SQLException.class)
    public void testCreateBuildSameNameNumberDate() throws SQLException {
        // TORE: [by fsi] use the error handling to verify we broke the correct constraint
        BuildEntity badBuild = new BuildEntity(1003L, "ba", "2", 1349002000000L, null, 1234L, "me", 0L, null, null);
        badBuild.setProperties(new HashSet<>());
        badBuild.setPromotions(new HashSet<>());
        saveBuild(badBuild);
    }

    @Test(expectedExceptions = SQLException.class)
    public void testCreateBuildSameId() throws SQLException {
        // TORE: [by fsi] use the error handling to verify we broke the correct constraint
        BuildEntity badBuild = new BuildEntity(1L, "bad", "1", 2L, null, 1234L, "me", 0L, null, null);
        badBuild.setProperties(new HashSet<>());
        badBuild.setPromotions(new HashSet<>());
        saveBuild(badBuild);
    }

    public void testFindBuildsForArtifactWithBuildRepo() throws SQLException {
        Collection<BuildEntity> builds = buildsDao.findBuildsForArtifactChecksum(ChecksumType.sha1,
                "acab88fc2a043c2479a6de676a2f8179e9ea2167", ALTERNATIVE_BUILD_INFO_REPO);

        assertEquals(builds.size(), 2);

        builds = buildsDao.findBuildsForArtifactChecksum(ChecksumType.sha1,
                "acab88fc2a043c2479a6de676a2f8179e9ea2167", NOT_EXISTING_BUILD_REPO);

        assertEquals(builds.size(), 0);
    }

    public void testFindBuildsForArtifactSHA1Checksum() throws SQLException {
        Collection<BuildEntity> builds = buildsDao.findBuildsForArtifactChecksum(ChecksumType.sha1,
                "acab88fc2a043c2479a6de676a2f8179e9ea2167", BuildAddon.BUILD_INFO_REPO_NAME);

        assertEquals(builds.size(), 3);
        Assertions.assertThat(builds).doesNotHaveDuplicates();
        for (BuildEntity build : builds) {
            assertTrue(build.getBuildId() == 1 || build.getBuildId() == 3 || build.getBuildId() == 7,
                    "Build id should be 1 or 3 or 7");
        }
    }

    public void testFindBuildsForArtifactMD5Checksum() throws SQLException {
        Collection<BuildEntity> builds = buildsDao.findBuildsForArtifactChecksum(ChecksumType.md5,
                "b02a360ecad98a34b59863c1e65bcf71", BuildAddon.BUILD_INFO_REPO_NAME);

        assertEquals(builds.size(), 3);
        Assertions.assertThat(builds).doesNotHaveDuplicates();
        for (BuildEntity build : builds) {
            assertTrue(build.getBuildId() == 1 || build.getBuildId() == 3 || build.getBuildId() == 7,
                    "Build id should be 1 or 3 or 7");
        }
    }

    public void testFindBuildsForDependencySHA1Checksum() throws SQLException {
        Collection<BuildEntity> builds = buildsDao.findBuildsForDependencyChecksum(ChecksumType.sha1,
                "ccab88fc2a043c2479a6de676a2f8179e9ea2167", BuildAddon.BUILD_INFO_REPO_NAME);

        assertEquals(builds.size(), 1);
        Assertions.assertThat(builds).doesNotHaveDuplicates();
        for (BuildEntity build : builds) {
            assertTrue(build.getBuildId() == 1 || build.getBuildId() == 2,
                    "Build id should be 1 or 2");
        }
    }

    public void testFindBuildsForDependencyWithBuildRepo() throws SQLException {
        Collection<BuildEntity> builds = buildsDao.findBuildsForDependencyChecksum(ChecksumType.sha1,
                "ccab88fc2a043c2479a6de676a2f8179e9ea2167", ALTERNATIVE_BUILD_INFO_REPO);

        assertEquals(builds.size(), 1);

        builds = buildsDao.findBuildsForDependencyChecksum(ChecksumType.sha1,
                "ccab88fc2a043c2479a6de676a2f8179e9ea2167", NOT_EXISTING_BUILD_REPO);

        assertEquals(builds.size(), 0);
    }

    public void testFindBuildsForDependencyMD5Checksum() throws SQLException {
        Collection<BuildEntity> builds = buildsDao.findBuildsForDependencyChecksum(ChecksumType.md5,
                "d02a360ecad98a34b59863c1e65bcf71", BuildAddon.BUILD_INFO_REPO_NAME);

        assertEquals(builds.size(), 2);
        Assertions.assertThat(builds).doesNotHaveDuplicates();
        for (BuildEntity build : builds) {
            assertTrue(build.getBuildId() == 7 || build.getBuildId() == 3, "Build id should be 2 or 3");
        }
    }

    private void doDelete(long buildId, int nbDeletedRows) throws SQLException {
        assertNotNull(buildsDao.getBuild(buildId));
        assertEquals(buildsDao.deleteBuild(buildId, false), nbDeletedRows);
        assertNull(buildsDao.getBuild(buildId));
    }

    private void assertAnyBuild(BuildEntity b) {
        assertNotNull(b);
        switch ((int) b.getBuildId()) {
            case 1:
                assertBuild1(b);
                break;
            case 2:
                assertBuild2(b);
                break;
            case 3:
                assertBuild3(b);
                break;
            case 4:
                assertBuild4(b);
                break;
            case 5:
                assertBuild5(b);
                break;
            case 6:
                assertBuild6(b);
                break;
            default:
                fail("Build with id " + b.getBuildId() + " is not known!");
        }
    }

    private void assertBuild1(BuildEntity b1) {
        assertEquals(b1.getBuildId(), 1L);
        assertEquals(b1.getBuildName(), "ba");
        assertEquals(b1.getBuildNumber(), "1");
        assertEquals(b1.getBuildDate(), 1349000000000L);
        assertNull(b1.getCiUrl());
        assertEquals(b1.getCreated(), 1350000000000L);
        assertEquals(b1.getCreatedBy(), "me");
        assertEquals(b1.getModified(), 1350000000001L);
        assertEquals(b1.getModifiedBy(), "not-me");
        ImmutableSet<BuildProperty> props = b1.getProperties();
        assertEquals(props.size(), 2);
        BuildProperty bp1 = new BuildProperty(1L, 1L, "start", "0");
        BuildProperty bp2 = new BuildProperty(2L, 1L, "status", "bad");
        assertTrue(props.contains(bp1));
        assertTrue(props.contains(bp2));
        for (BuildProperty prop : props) {
            if (prop.getPropId() == 1L) {
                assertTrue(prop.isIdentical(bp1));
            }
            if (prop.getPropId() == 2L) {
                assertTrue(prop.isIdentical(bp2));
            }
        }
        assertTrue(b1.getPromotions().isEmpty());
    }

    private void assertBuild2(BuildEntity b2) {
        assertEquals(b2.getBuildId(), 2L);
        assertEquals(b2.getBuildName(), "bb");
        assertEquals(b2.getBuildNumber(), "1");
        assertEquals(b2.getBuildDate(), 1349001000000L);
        assertEquals(b2.getCiUrl(), "http://myserver/jenkins/bb/1");
        assertEquals(b2.getCreated(), 1350001000000L);
        assertEquals(b2.getCreatedBy(), "me");
        assertEquals(b2.getModified(), 0L);
        assertNull(b2.getModifiedBy());
        ImmutableSet<BuildProperty> props = b2.getProperties();
        assertEquals(props.size(), 2);
        BuildProperty bp1 = new BuildProperty(3L, 2L, "start", "1");
        BuildProperty bp2 = new BuildProperty(4L, 2L, "status", "not-too-bad");
        assertTrue(props.contains(bp1));
        assertTrue(props.contains(bp2));
        for (BuildProperty prop : props) {
            if (prop.getPropId() == 1L) {
                assertTrue(prop.isIdentical(bp1));
            }
            if (prop.getPropId() == 2L) {
                assertTrue(prop.isIdentical(bp2));
            }
        }
        assertTrue(b2.getPromotions().isEmpty());
    }

    private void assertBuild3(BuildEntity b3) {
        assertEquals(b3.getBuildId(), 3L);
        assertEquals(b3.getBuildName(), "ba");
        assertEquals(b3.getBuildNumber(), "2");
        assertEquals(b3.getBuildDate(), 1349002000000L);
        assertNull(b3.getCiUrl());
        assertEquals(b3.getCreated(), 1350002000000L);
        assertEquals(b3.getCreatedBy(), "me");
        assertEquals(b3.getModified(), 0L);
        assertNull(b3.getModifiedBy());
        assertTrue(b3.getProperties().isEmpty());
        assertEquals(b3.getPromotions().size(), 1);
        BuildPromotionStatus promotionStatus = new BuildPromotionStatus(
                3L, 1350012000000L,
                "me", "dead", null, "bad stuff", null);
        BuildPromotionStatus found = b3.getPromotions().first();
        assertTrue(found.isIdentical(promotionStatus));
    }

    private void assertBuild4(BuildEntity b4) {
        assertEquals(b4.getBuildId(), 4L);
        assertEquals(b4.getBuildName(), "bb");
        assertEquals(b4.getBuildNumber(), "2");
        assertEquals(b4.getBuildDate(), 1349003000000L);
        assertEquals(b4.getCiUrl(), "http://myserver/jenkins/bb/2");
        assertEquals(b4.getCreated(), 1350003000000L);
        assertEquals(b4.getCreatedBy(), "me");
        assertEquals(b4.getModified(), 0L);
        assertNull(b4.getModifiedBy());
        assertTrue(b4.getProperties().isEmpty());
        ImmutableSortedSet<BuildPromotionStatus> promotions = b4.getPromotions();
        assertEquals(promotions.size(), 3);
        ImmutableSortedSet<BuildPromotionStatus> expectedStatus = ImmutableSortedSet.of(
                new BuildPromotionStatus(
                        4L, 1350003000000L,
                        null, "staging", null, null, null),
                new BuildPromotionStatus(
                        4L, 1350013000000L,
                        "promoter", "promoted", "qa-local", "sending to QA", "me"),
                new BuildPromotionStatus(
                        4L, 1350023000000L,
                        "tester", "rollback", "lost-local", "Refused by QA", null)
        );
        assertEquals(promotions, expectedStatus);
        UnmodifiableIterator<BuildPromotionStatus> it1 = expectedStatus.iterator();
        UnmodifiableIterator<BuildPromotionStatus> promosIt = promotions.iterator();
        for (int i = 0; i < 3; i++) {
            assertTrue(promosIt.next().isIdentical(it1.next()));
        }
    }

    private void assertBuild5(BuildEntity b5) {
        assertEquals(b5.getBuildId(), 5L);
        assertEquals(b5.getBuildName(), "ba");
        assertEquals(b5.getBuildNumber(), "3");
        assertEquals(b5.getBuildDate(), 1349004000000L);
        assertNull(b5.getCiUrl());
        assertEquals(b5.getCreated(), 1350004000000L);
        assertEquals(b5.getCreatedBy(), "me");
        assertEquals(b5.getModified(), 0L);
        assertNull(b5.getModifiedBy());
        ImmutableSet<BuildProperty> props = b5.getProperties();
        assertEquals(props.size(), 2);
        BuildProperty bp1 = new BuildProperty(5L, 5L, "start", "4");
        BuildProperty bp2 = new BuildProperty(6L, 5L, "status", "good");
        assertTrue(props.contains(bp1));
        assertTrue(props.contains(bp2));
        for (BuildProperty prop : props) {
            if (prop.getPropId() == 1L) {
                assertTrue(prop.isIdentical(bp1));
            }
            if (prop.getPropId() == 2L) {
                assertTrue(prop.isIdentical(bp2));
            }
        }
        ImmutableSortedSet<BuildPromotionStatus> promotions = b5.getPromotions();
        assertEquals(promotions.size(), 2);
        BuildPromotionStatus first = new BuildPromotionStatus(
                5L, 1350024000000L,
                null, "staging", null, null, "jenkins");
        BuildPromotionStatus second = new BuildPromotionStatus(
                5L, 1350034000000L,
                "promoter", PromotionStatus.RELEASED, "public", "Full release", "rel");
        assertTrue(promotions.first().isIdentical(first));
        assertTrue(promotions.last().isIdentical(second));
    }

    private void assertBuild6(BuildEntity b6) {
        assertEquals(b6.getBuildId(), 6L);
        assertEquals(b6.getBuildName(), "bb");
        assertEquals(b6.getBuildNumber(), "3");
        assertEquals(b6.getBuildDate(), 1349004000000L);
        assertEquals(b6.getCiUrl(), "http://myserver/jenkins/bb/3");
        assertEquals(b6.getCreated(), 1350004000000L);
        assertEquals(b6.getCreatedBy(), "me");
        assertEquals(b6.getModified(), 0L);
        assertNull(b6.getModifiedBy());
        assertTrue(b6.getProperties().isEmpty());
    }

    @AfterClass
    public void fullDelete() throws SQLException {
        assertEquals(buildArtifactsDao.deleteAllBuildArtifacts(), 6);
        assertEquals(buildDependenciesDao.deleteAllBuildDependencies(), 5);
        assertEquals(buildModulesDao.deleteAllBuildModules(), 4);
        assertEquals(buildsDao.deleteAllBuilds(false), 17);
        for (int i = 1; i <= 5; i++) {
            assertTrue(propertiesDao.deleteNodeProperties(i) > 1);
            assertTrue(nodesDao.delete(i));
        }
        assertEquals(binariesDao.deleteEntry("acab88fc2a043c2479a6de676a2f8179e9ea2167"), 1);
    }
}
